import pathlib
from pathlib import PosixPath
from typing import List

from rbx import config
from rbx.config import Artifact, Language, format_vars
from rbx.grading import steps
from rbx.grading.judge.sandbox import MERGE_STDERR, SandboxParams
from rbx.grading.steps import (
    GradingArtifacts,
    GradingFileInput,
    GradingFileOutput,
    TestcaseIO,
)
from rbx.schema import DumpedProblem, Problem


def build_formatted_command(
    command: str, problem: DumpedProblem, lang: Language
) -> str:
    return format_vars(
        command,
        **problem.get_vars(),
        file=lang.get_file(problem.code),
        submit_file=lang.get_submit_file(problem.code),
    )


def build_preprocess_commands(problem: DumpedProblem, lang: Language) -> List[str]:
    return [
        build_formatted_command(cmd, problem, lang) for cmd in (lang.preprocess or [])
    ]


def build_preprocess_sandbox_params() -> SandboxParams:
    params = SandboxParams(
        max_processes=None,
        preserve_env=True,
    )
    params.add_mapped_directory(pathlib.Path('/usr'))
    params.add_mapped_directory(pathlib.Path('/etc'))
    return params


def build_compile_grading_artifacts(
    problem: DumpedProblem, lang: Language
) -> GradingArtifacts:
    res = GradingArtifacts(root=PosixPath('.'))
    file = lang.get_file(problem.code)
    submit_file = lang.get_submit_file(problem.code)
    # Copy input file.
    res.inputs.append(GradingFileInput(src=PosixPath(file), dest=PosixPath(file)))
    # Copy output file.
    if lang.has_submit_file():
        res.outputs.append(
            GradingFileOutput(
                src=PosixPath(submit_file),
                dest=PosixPath(submit_file),
            )
        )
    # Copy other artifacts.
    for artifact_name, artifact_cfg in lang.artifacts.items():
        artifact_cfg = artifact_cfg or Artifact()
        artifact_path = format_vars(
            artifact_name, **problem.get_vars(), file=file, submit_file=submit_file
        )
        res.outputs.append(
            GradingFileOutput(
                src=PosixPath(artifact_path),
                dest=PosixPath(artifact_cfg.filename or artifact_path),
                optional=artifact_cfg.optional,
                executable=artifact_cfg.executable,
            )
        )

    return res


def build_run_sandbox_params(problem: Problem, has_input: bool) -> SandboxParams:
    params = SandboxParams()
    params.timeout = problem.timeLimit * 2
    params.wallclock_timeout = problem.timeLimit * 5
    params.address_space = problem.memoryLimit or 1024  # 1 GB
    params.set_stdall(
        stdin=PosixPath('stdin.txt') if has_input else None,
        stdout=PosixPath('stdout.txt'),
        stderr=MERGE_STDERR,
    )
    return params


def build_run_grading_artifacts(
    testcase: TestcaseIO, persist_root: pathlib.Path
) -> GradingArtifacts:
    res = GradingArtifacts(root=PosixPath('.'))
    res.inputs.append(
        GradingFileInput(
            src=testcase.input,
            dest=PosixPath('stdin.txt'),
        )
    )
    res.outputs.append(
        GradingFileOutput(
            src=PosixPath('stdout.txt'),
            dest=persist_root / f'stdout-{testcase.index}.txt',
            maxlen=steps.MAX_STDOUT_LEN,
        )
    )
    return res


def build_checker_compile_grading_artifacts(
    problem: DumpedProblem, persist_root: pathlib.Path
) -> GradingArtifacts:
    res = GradingArtifacts(root=PosixPath('.'))
    if not problem.checker:
        return res

    checker_path = PosixPath(problem.checker)
    if not checker_path.is_file():
        checker_path = config.get_builtin_checker(problem.checker)
    if not checker_path:
        return res

    res.inputs.append(GradingFileInput(src=checker_path, dest=PosixPath('checker.cpp')))
    testlib = config.get_testlib()
    if testlib.is_file():
        res.inputs.append(GradingFileInput(src=testlib, dest=PosixPath('testlib.h')))
    res.outputs.append(
        GradingFileOutput(
            src=PosixPath('checker'), dest=persist_root / 'checker', executable=True
        )
    )
    return res


def build_checker_run_grading_artifacts(
    problem: DumpedProblem, persist_root: pathlib.Path
) -> GradingArtifacts:
    res = GradingArtifacts(root=PosixPath('.'))
    if not problem.checker:
        return res
    res.inputs.append(
        GradingFileInput(
            src=persist_root / 'checker', dest=PosixPath('checker'), executable=True
        )
    )
    return res



from __future__ import generators

import collections
import dataclasses
import pathlib
import shutil
from collections.abc import Iterator
from typing import Dict, List, Optional, Set

import rich
import rich.live
import rich.table
from more_itertools import seekable
from pydantic import BaseModel

from rbx import console
from rbx.box import checkers, environment, package
from rbx.box.code import compile_item, run_item
from rbx.box.environment import EnvironmentSandbox, ExecutionConfig, VerificationLevel
from rbx.box.generators import generate_output_for_testcase, generate_standalone
from rbx.box.schema import (
    ExpectedOutcome,
    GeneratorCall,
    Solution,
    Testcase,
    TestcaseGroup,
)
from rbx.box.testcases import find_built_testcases
from rbx.grading.steps import (
    DigestOrDest,
    DigestOrSource,
    Evaluation,
    Outcome,
    TestcaseIO,
    TestcaseLog,
)
from rbx.utils import StatusProgress, model_to_yaml

StructuredEvaluation = Dict[str, Dict[str, List[Optional[Evaluation]]]]


class EvaluationItem(BaseModel):
    solution_index: int
    group_name: str
    testcase_index: int
    eval: Evaluation


class GroupSkeleton(BaseModel):
    name: str
    testcases: List[Testcase]


class SolutionReportSkeleton(BaseModel):
    solutions: List[Solution]
    groups: List[GroupSkeleton]
    group_first: bool

    def find_group_skeleton(self, group_name: str) -> Optional[GroupSkeleton]:
        groups = [group for group in self.groups if group.name == group_name]
        if not groups:
            return None
        return groups[0]

    def empty_structured_evaluation(self) -> StructuredEvaluation:
        res: StructuredEvaluation = {}
        for solution in self.solutions:
            res[str(solution.path)] = {}
            for group in self.groups:
                res[str(solution.path)][group.name] = [None for _ in group.testcases]
        return res


@dataclasses.dataclass
class RunSolutionResult:
    skeleton: SolutionReportSkeleton
    items: Iterator[EvaluationItem]

    def empty_structured_evaluation(self) -> StructuredEvaluation:
        return self.skeleton.empty_structured_evaluation()


def is_fast(solution: Solution) -> bool:
    # If solution has TLE tag, it is considered slow.
    return not solution.outcome.match(Outcome.TIME_LIMIT_EXCEEDED)


def get_matching_solutions(expected_outcome: ExpectedOutcome) -> List[Solution]:
    res = []
    for solution in package.get_solutions():
        if not solution.outcome.intersect(expected_outcome):
            continue
        res.append(solution)
    return res


def compile_solutions(
    progress: Optional[StatusProgress] = None,
    tracked_solutions: Optional[Set[str]] = None,
) -> Dict[pathlib.Path, str]:
    pkg = package.find_problem_package_or_die()

    compiled_solutions = {}

    for solution in pkg.solutions:
        if (
            tracked_solutions is not None
            and str(solution.path) not in tracked_solutions
        ):
            continue
        if progress:
            progress.update(f'Compiling solution [item]{solution.path}[/item]...')
        try:
            compiled_solutions[solution.path] = compile_item(solution)
        except:
            console.console.print(
                f'[error]Failed compiling solution [item]{solution.path}[/item][/error]'
            )
            raise

    return compiled_solutions


def _run_solution_on_testcase(
    solution: Solution,
    compiled_digest: str,
    checker_digest: Optional[str],
    testcase: Testcase,
    output_dir: pathlib.Path,
    testcase_index: int = 0,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> Evaluation:
    pkg = package.find_problem_package_or_die()
    actual_sandbox = package.get_singleton_sandbox()

    timelimit = pkg.timelimit_for_language(solution.language)

    sandbox = EnvironmentSandbox()
    sandbox.timeLimit = timelimit
    if verification.value >= VerificationLevel.FULL.value:
        # Use double TL.
        sandbox.timeLimit = sandbox.timeLimit * 2
    sandbox.wallTimeLimit = (
        timelimit * 2 if actual_sandbox.use_soft_timeout() else sandbox.timeLimit
    )
    sandbox.memoryLimit = pkg.memorylimit_for_language(solution.language)
    sandbox.fileSizeLimit = pkg.outputLimit
    extra_config = ExecutionConfig(sandbox=sandbox)

    output_path = output_dir / testcase.inputPath.with_suffix('.out').name
    error_path = output_path.with_suffix('.err')
    log_path = output_path.with_suffix('.log')
    output_path.parent.mkdir(parents=True, exist_ok=True)

    run_log = run_item(
        solution,
        DigestOrSource.create(compiled_digest),
        stdin=DigestOrSource.create(testcase.inputPath),
        stdout=DigestOrDest.create(output_path),
        stderr=DigestOrDest.create(error_path),
        extra_config=extra_config,
    )

    if checker_digest is not None:
        checker_result = checkers.check(
            checker_digest,
            run_log,
            testcase,
            program_output=output_path,
        )
    else:
        checker_result = checkers.check_with_no_output(run_log)

    eval = Evaluation(
        result=checker_result,
        testcase=TestcaseIO(
            index=testcase_index, input=testcase.inputPath, output=testcase.outputPath
        ),
        log=TestcaseLog(
            **(run_log.model_dump() if run_log is not None else {}),
            stdout_absolute_path=output_path.absolute(),
            stderr_absolute_path=error_path.absolute(),
            log_absolute_path=log_path.absolute(),
        ),
    )

    log_path.write_text(model_to_yaml(eval))
    return eval


def _run_solution(
    solution: Solution,
    compiled_digest: str,
    checker_digest: Optional[str],
    solution_index: int,
    group_name: str,
    progress: Optional[StatusProgress] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> Iterator[Evaluation]:
    runs_dir = package.get_problem_runs_dir()

    group = package.get_testgroup(group_name)
    testcases = find_built_testcases(group)
    for i, testcase in enumerate(testcases):
        assert testcase.outputPath is not None
        output_path = runs_dir / f'{solution_index}' / group.name

        if progress:
            progress.update(
                f'Running solution [item]{solution.path}[/item] on test [item]{group.name}[/item] / [item]{i}[/item]...'
            )

        yield _run_solution_on_testcase(
            solution,
            compiled_digest,
            checker_digest,
            testcase,
            output_path,
            testcase_index=i,
            verification=verification,
        )


def convert_list_of_solution_evaluations_to_dict(
    items: Iterator[EvaluationItem],
) -> List[Dict[str, List[Evaluation]]]:
    pkg = package.find_problem_package_or_die()
    res: List[Dict[str, List[Evaluation]]] = [
        collections.defaultdict(list) for _ in pkg.solutions
    ]

    for item in items:
        res[item.solution_index][item.group_name].append(item.eval)

    return res


def _get_report_skeleton(
    tracked_solutions: Optional[Set[str]] = None,
    group_first: bool = False,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> SolutionReportSkeleton:
    pkg = package.find_problem_package_or_die()
    solutions = [
        sol
        for sol in pkg.solutions
        if verification.value >= VerificationLevel.ALL_SOLUTIONS.value or is_fast(sol)
    ]
    if tracked_solutions is not None:
        solutions = [
            solution
            for solution in solutions
            if str(solution.path) in tracked_solutions
        ]

    groups = []
    for group in pkg.testcases:
        testcases = find_built_testcases(group)
        groups.append(GroupSkeleton(name=group.name, testcases=testcases))
    return SolutionReportSkeleton(
        solutions=solutions, groups=groups, group_first=group_first
    )


def _produce_solution_items(
    progress: Optional[StatusProgress] = None,
    tracked_solutions: Optional[Set[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    check: bool = True,
    group_first: bool = False,
) -> Iterator[EvaluationItem]:
    pkg = package.find_problem_package_or_die()

    checker_digest = checkers.compile_checker() if check else None
    compiled_solutions = compile_solutions(
        progress=progress, tracked_solutions=tracked_solutions
    )

    # Clear run directory and rely on cache to
    # repopulate it.
    runs_dir = package.get_problem_runs_dir()
    shutil.rmtree(str(runs_dir), ignore_errors=True)
    runs_dir.mkdir(parents=True, exist_ok=True)
    solutions = list(
        (i, sol)
        for i, sol in enumerate(pkg.solutions)
        if verification.value >= VerificationLevel.ALL_SOLUTIONS.value or is_fast(sol)
    )
    if tracked_solutions is not None:
        solutions = [
            (i, sol) for i, sol in solutions if str(sol.path) in tracked_solutions
        ]

    def yield_items(
        solution_index: int, solution: Solution, group_name: str
    ) -> Iterator[EvaluationItem]:
        for i, eval in enumerate(
            _run_solution(
                solution,
                compiled_solutions[solution.path],
                checker_digest,
                solution_index,
                group_name,
                progress=progress,
                verification=verification,
            )
        ):
            yield EvaluationItem(
                solution_index=solution_index,
                group_name=group_name,
                testcase_index=i,
                eval=eval,
            )

    groups = pkg.testcases
    if group_first:
        for group in groups:
            for i, solution in solutions:
                yield from yield_items(i, solution, group.name)
        return

    for i, solution in solutions:
        for group in groups:
            yield from yield_items(i, solution, group.name)


def run_solutions(
    progress: Optional[StatusProgress] = None,
    tracked_solutions: Optional[Set[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    check: bool = True,
    group_first: bool = False,
) -> RunSolutionResult:
    return RunSolutionResult(
        skeleton=_get_report_skeleton(
            tracked_solutions, group_first, verification=verification
        ),
        items=_produce_solution_items(
            progress=progress,
            tracked_solutions=tracked_solutions,
            verification=verification,
            check=check,
            group_first=group_first,
        ),
    )


def _run_interactive_solutions(
    tracked_solutions: Optional[Set[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    generator: Optional[GeneratorCall] = None,
    check: bool = True,
) -> Iterator[EvaluationItem]:
    pkg = package.find_problem_package_or_die()
    main_solution = package.get_main_solution()
    check = check and main_solution is not None

    checker_digest = checkers.compile_checker() if check else None
    compiled_solutions = compile_solutions(tracked_solutions=tracked_solutions)

    main_solution_digest = None
    if check and main_solution is not None:
        try:
            main_solution_digest = compile_item(main_solution)
        except:
            console.console.print(
                '[error]Failed compiling main solution. If you do not want to check against a main solution, run with --nocheck flag.[/error]'
            )
            raise

    solutions = list(enumerate(pkg.solutions))
    if tracked_solutions is not None:
        solutions = [
            (i, sol) for i, sol in solutions if str(sol.path) in tracked_solutions
        ]

    irun_dir = package.get_problem_runs_dir() / '.irun'
    shutil.rmtree(str(irun_dir), ignore_errors=True)
    irun_dir.mkdir(parents=True, exist_ok=True)
    inputs_dir = irun_dir / 'inputs'
    inputs_dir.mkdir(parents=True, exist_ok=True)
    input_path = inputs_dir / '000.in'
    output_path = input_path.with_suffix('.out')

    if generator is not None:
        expanded_call = generate_standalone(generator, input_path)
        console.console.print(
            f'Using input from generator call [item]{expanded_call.name} {expanded_call.args}[/item].'
        )
    else:
        input = console.multiline_prompt('Testcase input')
        input_path.write_text(input)
    testcase = Testcase(inputPath=input_path, outputPath=output_path if check else None)

    if main_solution_digest is not None:
        # TODO: Add stderr path
        generate_output_for_testcase(main_solution_digest, testcase)

    for i, solution in solutions:
        output_dir = irun_dir / f'{i}'

        yield EvaluationItem(
            solution_index=i,
            group_name='irun',
            testcase_index=0,
            eval=_run_solution_on_testcase(
                solution,
                compiled_solutions[solution.path],
                checker_digest,
                testcase,
                output_dir,
                verification=verification,
            ),
        )


def run_and_print_interactive_solutions(
    tracked_solutions: Optional[Set[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
    generator: Optional[GeneratorCall] = None,
    check: bool = True,
    print: bool = False,
):
    pkg = package.find_problem_package_or_die()
    items = _run_interactive_solutions(
        tracked_solutions=tracked_solutions,
        verification=verification,
        check=check,
        generator=generator,
    )

    for item in items:
        sol = pkg.solutions[item.solution_index]
        _print_solution_header(sol, console.console)

        stdout_path = item.eval.log.stdout_absolute_path
        if print:
            if (
                item.eval.testcase.output is not None
                and stdout_path is not None
                and stdout_path.is_file()
            ):
                console.console.print(stdout_path.read_text())
            else:
                console.console.print('[warning]Solution produced no output.[/warning]')
        elif stdout_path is not None:
            console.console.print(f'Output: {stdout_path}.')
            console.console.print()


def get_outcome_style_verdict(outcome: Outcome) -> str:
    if outcome == Outcome.ACCEPTED:
        return 'green'
    if outcome == Outcome.WRONG_ANSWER:
        return 'red'
    if outcome == Outcome.TIME_LIMIT_EXCEEDED:
        return 'yellow'
    if outcome == Outcome.RUNTIME_ERROR:
        return 'lnumber'
    if outcome == Outcome.MEMORY_LIMIT_EXCEEDED:
        return 'cyan'
    return 'magenta'


def get_testcase_markup_verdict(eval: Evaluation) -> str:
    res = '✓'
    if eval.result.outcome != Outcome.ACCEPTED:
        res = '✗'
    if eval.result.outcome == Outcome.TIME_LIMIT_EXCEEDED:
        res = '⧖'
    if eval.result.outcome == Outcome.RUNTIME_ERROR:
        res = '✗'
    style = get_outcome_style_verdict(eval.result.outcome)
    res = f'[{style}]{res}[/{style}]'
    # if eval.log.stdout_absolute_path:
    #     output_path = eval.log.stdout_absolute_path.resolve()
    #     output_link = f'file://{output_path}'
    #     res = f'[link={output_link}]{res}[/link]'
    return res


def _get_evals_time_in_ms(evals: List[Evaluation]) -> int:
    if not evals:
        return 0
    return max(int((eval.log.time or 0.0) * 1000) for eval in evals)


def _get_evals_memory_in_mb(evals: List[Evaluation]) -> int:
    if not evals:
        return 0
    return max(int(eval.log.memory or 0) // (1024 * 1024) for eval in evals)


def _get_evals_memory_in_bytes(evals: List[Evaluation]) -> int:
    if not evals:
        return 0
    return max(int(eval.log.memory or 0) for eval in evals)


def get_evals_formatted_time(evals: List[Evaluation]) -> str:
    max_time = _get_evals_time_in_ms(evals)
    return f'{max_time} ms'


def get_evals_formatted_memory(evals: List[Evaluation]) -> str:
    max_memory = _get_evals_memory_in_bytes(evals)
    if max_memory < 1024 * 1024:
        if max_memory < 1024:
            return f'{max_memory} B'
        return f'{max_memory // 1024} KiB'
    return f'{max_memory // (1024 * 1024)} MiB'


def _print_solution_outcome(
    solution: Solution,
    evals: List[Evaluation],
    console: rich.console.Console,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> bool:
    pkg = package.find_problem_package_or_die()

    has_plain_tle = False
    bad_verdicts = set()
    no_tle_bad_verdicts = set()
    for eval in evals:
        if eval.result.outcome != Outcome.ACCEPTED:
            bad_verdicts.add(eval.result.outcome)
        if (
            eval.result.no_tle_outcome is not None
            and eval.result.no_tle_outcome != Outcome.ACCEPTED
        ):
            no_tle_bad_verdicts.add(eval.result.no_tle_outcome)
        has_plain_tle = has_plain_tle or (
            eval.result.outcome == Outcome.TIME_LIMIT_EXCEEDED
            and eval.result.no_tle_outcome is None
        )

    unmatched_bad_verdicts = set(
        v for v in bad_verdicts if not solution.outcome.match(v)
    )
    matched_bad_verdicts = bad_verdicts - unmatched_bad_verdicts
    expected_outcome_is_bad = not solution.outcome.match(Outcome.ACCEPTED)

    if unmatched_bad_verdicts or (expected_outcome_is_bad and not matched_bad_verdicts):
        console.print('[error]FAILED[/error]', end=' ')
    else:
        console.print('[success]OK[/success]', end=' ')

    console.print(f'Expected: {solution.outcome}', end='')

    if unmatched_bad_verdicts:
        unmatched_bad_verdicts_names = set(v.name for v in unmatched_bad_verdicts)
        console.print(f', got: {" ".join(unmatched_bad_verdicts_names)}', end='')
    elif expected_outcome_is_bad and not matched_bad_verdicts:
        console.print(f', got: {Outcome.ACCEPTED.name}', end='')

    console.print()
    evals_time = _get_evals_time_in_ms(evals)
    expected_outcome_is_tle = solution.outcome.match(Outcome.TIME_LIMIT_EXCEEDED)
    if (
        # Running verification with double TL.
        verification.value >= VerificationLevel.FULL.value
        # Solution expects a TLE.
        and expected_outcome_is_tle
        # Solution does not have a plain TLE.
        and not has_plain_tle
        # A TLE has happened.
        and Outcome.TIME_LIMIT_EXCEEDED in matched_bad_verdicts
        # The solution runs in double TL.
        and evals_time < pkg.timelimit_for_language(solution.language) * 2
    ):
        other_verdicts = (bad_verdicts | no_tle_bad_verdicts) - {
            Outcome.TIME_LIMIT_EXCEEDED
        }
        if not other_verdicts:
            # The solution has no other bad verdicts except for TLEs in double TL.
            console.print(
                '[yellow]WARNING[/yellow] The solution still passed in double TL.'
            )
        elif not (bad_verdicts - {Outcome.TIME_LIMIT_EXCEEDED}):
            # The solution has other bad soft TLE outcomes.
            other_verdicts_names = ' '.join(v.name for v in other_verdicts)
            console.print(
                f'[yellow]WARNING[/yellow] The solution could still run under double TL, but failed with [item]{other_verdicts_names}[/item].'
            )
    console.print(f'Time: {get_evals_formatted_time(evals)}')
    console.print(f'Memory: {get_evals_formatted_memory(evals)}')
    return len(unmatched_bad_verdicts) == 0


def _consume_and_key_evaluation_items(
    items: Iterator[EvaluationItem],
    skeleton: SolutionReportSkeleton,
) -> Iterator[StructuredEvaluation]:
    """
    Consumes EvaluationItems from a run_solutions call and build a view
    with them, possibly marking with optional unprocessed items.
    """
    pkg = package.find_problem_package_or_die()
    res = skeleton.empty_structured_evaluation()

    for item in items:
        solution = pkg.solutions[item.solution_index]
        res[str(solution.path)][item.group_name][item.testcase_index] = item.eval
        yield res


def _print_solution_header(solution: Solution, console: rich.console.Console):
    solutions = package.get_solutions()
    solution_index = [
        i for i, sol in enumerate(solutions) if sol.path == solution.path
    ][0]
    solution_testdir = package.get_problem_runs_dir() / f'{solution_index}'
    console.print(f'[item]{solution.path}[/item]', end=' ')
    console.print(f'({solution_testdir})')


def _print_timing(
    console: rich.console.Console,
    skeleton: SolutionReportSkeleton,
    structured_evaluation: StructuredEvaluation,
):
    slowest_good = None
    fastest_slow = None
    for solution in skeleton.solutions:
        evals_per_group = structured_evaluation[str(solution.path)]
        all_evals = []
        for evals in evals_per_group.values():
            all_evals.extend([eval for eval in evals if eval is not None])
        solution_time = _get_evals_time_in_ms(all_evals)
        if solution.outcome.match(Outcome.ACCEPTED):
            if slowest_good is None or solution_time > slowest_good:
                slowest_good = solution_time
        if solution.outcome.is_slow():
            if fastest_slow is None or solution_time < fastest_slow:
                fastest_slow = solution_time

    if slowest_good is None and fastest_slow is None:
        return

    console.print('[status]Timing summary:[/status]')
    if slowest_good is not None:
        console.print(f'Slowest [success]OK[/success] solution: {slowest_good} ms')
    if fastest_slow is not None:
        console.print(f'Fastest [error]slow[/error] solution: {fastest_slow} ms')


def _render_detailed_group_table(
    group: TestcaseGroup,
    skeleton: SolutionReportSkeleton,
    structured_evaluations: Iterator[StructuredEvaluation],
    console: rich.console.Console,
):
    group_skeleton = skeleton.find_group_skeleton(group.name)
    assert group_skeleton is not None

    def generate_table(
        structured_evaluation: StructuredEvaluation, group_name: str
    ) -> rich.table.Table:
        table = rich.table.Table()
        for solution in skeleton.solutions:
            table.add_column(f'[item]{solution.path}[/item]', justify='full')

        evals_per_solution = collections.defaultdict(list)
        for tc, _ in enumerate(group_skeleton.testcases):
            row = []
            for solution in skeleton.solutions:
                eval = structured_evaluation[str(solution.path)][group_name][tc]
                evals_per_solution[str(solution.path)].append(eval)
                if eval is None:
                    row.append('...')
                    continue
                verdict = get_testcase_markup_verdict(eval)
                time = get_evals_formatted_time([eval])
                row.append(f'{verdict} {time}')
            table.add_row(*row)

        if table.row_count > 0:
            summary_row = []
            for solution in skeleton.solutions:
                evals = evals_per_solution[str(solution.path)]
                non_null_evals = [eval for eval in evals if eval is not None]
                if not non_null_evals:
                    summary_row.append('...')
                    continue
                summary_row.append('  ' + get_evals_formatted_time(non_null_evals))
            table.add_section()
            table.add_row(*summary_row)
        return table

    with rich.live.Live(
        generate_table(skeleton.empty_structured_evaluation(), group.name),
        refresh_per_second=5,
        console=console,
    ) as live:
        for _ in skeleton.solutions:
            for _ in group_skeleton.testcases:
                structured_evaluation = next(structured_evaluations)
                live.update(generate_table(structured_evaluation, group.name))
                live.refresh()


def _print_detailed_run_report(
    result: RunSolutionResult,
    console: rich.console.Console,
    structured_evaluations: Iterator[StructuredEvaluation],
    timing: bool = True,
):
    structured_evaluations = seekable(structured_evaluations)
    for group in result.skeleton.groups:
        console.print(f'[bold][status]{group.name}[/status][/bold]')

        _render_detailed_group_table(
            package.get_testgroup(group.name),
            result.skeleton,
            structured_evaluations,
            console,
        )
        continue

    ok = True
    structured_evaluations.seek(-1)
    structured_evaluation = next(structured_evaluations)
    for solution in result.skeleton.solutions:
        all_evals = []
        for evals in structured_evaluation[str(solution.path)].values():
            all_evals.extend(evals)
        _print_solution_header(solution, console)
        cur_ok = _print_solution_outcome(
            solution,
            all_evals,
            console,
        )
        ok = ok and cur_ok
        console.print()

    console.print()

    if timing:
        _print_timing(console, result.skeleton, structured_evaluation)
    return ok


def print_run_report(
    result: RunSolutionResult,
    console: rich.console.Console,
    verification: environment.VerificationParam,
    detailed: bool = False,
    timing: bool = True,
) -> bool:
    pkg = package.find_problem_package_or_die()
    items = seekable(result.items)
    structured_evaluations = _consume_and_key_evaluation_items(items, result.skeleton)
    if detailed:
        return _print_detailed_run_report(
            result, console, structured_evaluations, timing=timing
        )

    assert not result.skeleton.group_first
    # Since we're now streaming the evaluation results, the for-loop is a bit
    # confusing. We must keep state across the iteration to understand whether
    # we're seeing a new solution or a new testgroup.
    ok = True
    last_solution: Optional[Solution] = None
    last_group: Optional[str] = None
    test_index = 0
    all_evals = []
    group_evals = []

    def print_last_solution():
        nonlocal ok
        if last_solution is None:
            return
        cur_ok = _print_solution_outcome(
            last_solution,
            all_evals,
            console,
            verification=VerificationLevel(verification),
        )
        console.print()
        ok = ok and cur_ok

    for item in items:
        eval = item.eval
        solution = pkg.solutions[item.solution_index]
        is_new_solution = last_solution is None or solution.path != last_solution.path
        is_new_group = is_new_solution or last_group != item.group_name
        is_closing_group = last_group is not None and is_new_group

        if is_closing_group:
            console.print(
                f'({get_evals_formatted_time(group_evals)}, {get_evals_formatted_memory(group_evals)})',
                end='',
            )
            console.print()

        if is_new_solution:
            print_last_solution()
            all_evals = []
            last_solution = solution
            _print_solution_header(last_solution, console)

        if is_new_group:
            group_evals = []
            last_group = item.group_name
            test_index = 0
            console.print(f'[bold][status]{item.group_name}[/status][/bold]', end=' ')

        all_evals.append(eval)
        group_evals.append(eval)
        console.print(f'{test_index}/', end='')
        console.print(get_testcase_markup_verdict(eval), end=' ')

        test_index += 1

    console.print(
        f'({get_evals_formatted_time(group_evals)}, {get_evals_formatted_memory(group_evals)})',
        end=' ',
    )
    console.print()
    print_last_solution()

    items.seek(0)
    structured_evaluations_list = list(structured_evaluations)

    if structured_evaluations_list:
        _print_timing(console, result.skeleton, structured_evaluations_list[-1])

    return ok



from typing import Optional

from pydantic import BaseModel, Field

from rbx.box.packaging.boca.extension import BocaExtension, BocaLanguageExtension


# List of extensions defined in-place.
class MacExtension(BaseModel):
    gpp_alternative: Optional[str] = None


# Extension abstractions.
class Extensions(BaseModel):
    mac: Optional[MacExtension] = Field(
        None, description='Extension for setting mac-only configuration.'
    )
    boca: Optional[BocaExtension] = Field(
        None, description='Environment-level extensions for BOCA packaging.'
    )


class LanguageExtensions(BaseModel):
    boca: Optional[BocaLanguageExtension] = Field(
        None, description='Language-level extensions for BOCA packaging.'
    )



import abc
import pathlib
from typing import Any

from rbx.schema import Problem


class Submitor(abc.ABC):
    @abc.abstractmethod
    def key(self) -> str:
        pass

    @abc.abstractmethod
    def should_handle(self, problem: Problem) -> bool:
        pass

    @abc.abstractmethod
    def submit(
        self,
        file: pathlib.Path,
        problem: Problem,
        submitor_config: Any,
        credentials: Any,
    ) -> bool:
        pass



import functools
import pathlib
from typing import Optional

import typer

from rbx import console, utils


def find_package(root: pathlib.Path = pathlib.Path()) -> Optional[pathlib.Path]:
    root = root.resolve()

    def has_file():
        problem_yaml_path = root / 'problem.rbx.yml'
        contest_yaml_path = root / 'contest.rbx.yml'
        return problem_yaml_path.is_file() or contest_yaml_path.is_file()

    while root != pathlib.PosixPath('/') and not has_file():
        root = root.parent
    if not has_file():
        return None
    return root


def within_closest_package(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        package = find_package()
        if package is None:
            console.console.print('[error]No rbx package found.[/error]')
            raise typer.Exit(1)
        # Get deepest package.
        with utils.new_cd(package):
            return func(*args, **kwargs)

    return wrapper



import pathlib

from rbx.config import Language, get_config
from rbx.schema import Problem
from rbx.submitors.codeforces import CodeforcesSubmitor

_SUBMITORS = [CodeforcesSubmitor()]


def handle_submit(file: pathlib.Path, problem: Problem, lang: Language) -> bool:
    for submitor in _SUBMITORS:
        if submitor.should_handle(problem):
            submitor_config = get_config().submitor[lang.submitor]
            credentials = get_config().credentials[submitor.key()]
            return submitor.submit(
                file, problem, submitor_config[submitor.key()], credentials
            )
    return False



import uuid
from typing import Dict, List, Optional

from pydantic import BaseModel

from rbx import utils


class Testcase(BaseModel):
    input: str
    output: str


class Batch(BaseModel):
    id: str
    size: int

    @staticmethod
    def create():
        return Batch(id=str(uuid.uuid4()), size=1)


class Problem(BaseModel):
    name: str
    group: str = ''
    url: str = ''
    interactive: bool = False
    memoryLimit: int
    timeLimit: int
    tests: List[Testcase] = []
    testType: str = 'single'
    batch: Batch

    def get_code(self):
        return self.get_normalized_name()

    def get_normalized_name(self) -> str:
        return utils.normalize_with_underscores(self.name)


class DumpedProblem(Problem):
    code: str
    aliases: List[str]
    checker: Optional[str] = None

    @staticmethod
    def from_problem(problem: Problem, **kwargs) -> 'DumpedProblem':
        return DumpedProblem(**problem.model_dump(), **kwargs)

    def pretty_name(self) -> str:
        if self.name == self.code:
            return self.name
        return f'{self.name} ({self.code})'

    def get_vars(self) -> Dict[str, str]:
        return {
            'problem_name': self.name,
            'problem_code': self.code,
            'problem_url': self.url,
            'problem_contest': self.group,
            'problem_time_limit': f'{self.timeLimit}ms',
            'problem_memory_limit': f'{self.memoryLimit}MB',
            'problem_test_type': self.testType,
        }



from __future__ import annotations

import os
import pathlib
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, model_validator
from pydantic_core import PydanticCustomError

from rbx.autoenum import AutoEnum, alias
from rbx.box.statements.schema import Statement
from rbx.grading.steps import Outcome

Primitive = Union[str, int, float, bool]


def NameField(**kwargs):
    return Field(
        pattern=r'^[a-zA-Z0-9][a-zA-Z0-9\-_]*$', min_length=3, max_length=32, **kwargs
    )


def _check_oneof(model_obj: BaseModel, fields: List[str]):
    has = []
    for field in fields:
        if hasattr(model_obj, field) and getattr(model_obj, field):
            has.append(field)
    if len(has) <= 1:
        return
    raise ValueError(
        f'fields {has} were specified at the same time '
        'in a testgroup; only one of them can be specified'
    )


def expand_var(value: Primitive) -> Primitive:
    if not isinstance(value, str):
        return value
    if value.startswith('\\'):
        return value[1:]
    if not value.startswith('py`') or not value.endswith('`'):
        return value
    res = eval(value[3:-1])
    for supported_type in [str, int, float, bool]:
        if isinstance(res, supported_type):
            return res

    raise TypeError(
        f'Variable with backticks should evaluate to a primitive Python type: {value}'
    )


class ExpectedOutcome(AutoEnum):
    ACCEPTED = alias('accepted', 'ac', 'correct')  # type: ignore
    """Expected outcome for correct solutions (AC)."""

    WRONG_ANSWER = alias('wrong answer', 'wa')  # type: ignore
    """Expected outcome for solutions that finish successfully,
    but the produced output are incorrect (WA)."""

    INCORRECT = alias('fail', 'incorrect')  # type: ignore
    """Expected outcome for solutions that finish with any non-AC verdict."""

    RUNTIME_ERROR = alias('runtime error', 'rte', 're')  # type: ignore
    """Expected outcome solutions that finish with non-zero code (RTE)."""

    TIME_LIMIT_EXCEEDED = alias('time limit exceeded', 'timeout', 'tle')  # type: ignore
    """Expected outcome for solutions that do not finish in time."""

    MEMORY_LIMIT_EXCEEDED = alias('memory limit exceeded', 'mle')  # type: ignore
    """Expected outcome for solutions that use more memory than allowed."""

    OUTPUT_LIMIT_EXCEEDED = alias('output limit exceeded', 'ole')  # type: ignore
    """Expected outcome for solutions that use more output than allowed."""

    TLE_OR_RTE = alias('tle or rte', 'tle/rte', 'tle+rte')  # type: ignore
    """Expected outcome for solutions that finish with either TLE or RTE.
    
    Especially useful for environments where TLE and RTE are indistinguishable."""

    def style(self) -> str:
        if self == ExpectedOutcome.ACCEPTED:
            return 'green'
        if self == ExpectedOutcome.WRONG_ANSWER:
            return 'red'
        if self == ExpectedOutcome.INCORRECT:
            return 'red'
        if self.match(Outcome.TIME_LIMIT_EXCEEDED):
            return 'yellow'
        if self.match(Outcome.RUNTIME_ERROR):
            return 'lnumber'
        if self.match(Outcome.MEMORY_LIMIT_EXCEEDED):
            return 'cyan'
        return 'magenta'

    def is_slow(self) -> bool:
        return self in [ExpectedOutcome.TIME_LIMIT_EXCEEDED, ExpectedOutcome.TLE_OR_RTE]

    def match(self, outcome: Outcome) -> bool:
        if self == ExpectedOutcome.ACCEPTED:
            return outcome == Outcome.ACCEPTED
        if self == ExpectedOutcome.WRONG_ANSWER:
            return outcome == Outcome.WRONG_ANSWER
        if self == ExpectedOutcome.INCORRECT:
            return outcome in {
                Outcome.WRONG_ANSWER,
                Outcome.RUNTIME_ERROR,
                Outcome.MEMORY_LIMIT_EXCEEDED,
                Outcome.TIME_LIMIT_EXCEEDED,
                Outcome.OUTPUT_LIMIT_EXCEEDED,
            }
        if self == ExpectedOutcome.RUNTIME_ERROR:
            return outcome == Outcome.RUNTIME_ERROR
        if self == ExpectedOutcome.TIME_LIMIT_EXCEEDED:
            return outcome == Outcome.TIME_LIMIT_EXCEEDED
        if self == ExpectedOutcome.MEMORY_LIMIT_EXCEEDED:
            return outcome == Outcome.MEMORY_LIMIT_EXCEEDED
        if self == ExpectedOutcome.TLE_OR_RTE:
            return outcome in {Outcome.TIME_LIMIT_EXCEEDED, Outcome.RUNTIME_ERROR}
        if self == ExpectedOutcome.OUTPUT_LIMIT_EXCEEDED:
            return outcome == Outcome.OUTPUT_LIMIT_EXCEEDED
        return False

    def get_matches(self) -> List[Outcome]:
        return [outcome for outcome in Outcome if self.match(outcome)]

    def intersect(self, rhs: 'ExpectedOutcome') -> bool:
        return bool(set(self.get_matches()) & set(rhs.get_matches()))


class CodeItem(BaseModel):
    model_config = ConfigDict(extra='forbid')

    path: pathlib.Path = Field(
        description="""The path to the code file, relative to the package directory."""
    )

    language: Optional[str] = Field(
        None, description="""The language of the code file."""
    )

    compilationFiles: Optional[List[str]] = Field(
        [],
        description="""
Extra files that should be placed alongside the code file during its compilation,
such as testlib.h, jngen.h, etc.

The paths should be given relative to the package directory, but will be included
relative to the `path` directory.

Testlib and jngen are already included by default.
""",
    )


class Testcase(BaseModel):
    model_config = ConfigDict(extra='forbid')

    inputPath: pathlib.Path = Field(description="""The path of the input file.""")

    outputPath: Optional[pathlib.Path] = Field(
        None, description="""The path of the output file."""
    )


class GeneratorCall(BaseModel):
    model_config = ConfigDict(extra='forbid')

    name: str = NameField(description='The name of the generator to call.')

    args: Optional[str] = Field(
        None, description='The arguments to pass to the generator.'
    )


class TestcaseSubgroup(BaseModel):
    model_config = ConfigDict(extra='forbid')

    name: str = NameField(description='The name of the test group.')

    testcases: List[Testcase] = Field(
        [],
        description="""
The path of testcases to add to this group,
in the order they're defined.""",
    )

    testcaseGlob: Optional[str] = Field(
        None,
        description="""
A Python glob that matches input file paths relative to the
package directory. The globbed files should end with the extension
".in", and their corresponding outputs, if defined, should have the same file name,
but ending with ".out".
""",
    )

    generators: List[GeneratorCall] = Field(
        [],
        description="""
A list of generators to call to generate testcases for this group.
""",
    )

    generatorScript: Optional[CodeItem] = Field(
        None,
        description="""
A generator script to call to generate testcases for this group.
""",
    )

    @model_validator(mode='after')
    def check_oneof(self) -> 'TestcaseSubgroup':
        _check_oneof(
            self,
            [
                'testcases',
                'testcaseGlob',
                'generators',
                'generatorScript',
            ],
        )
        return self


class TestcaseGroup(TestcaseSubgroup):
    model_config = ConfigDict(extra='forbid')

    subgroups: List[TestcaseSubgroup] = Field(
        [],
        description="""
A list of test subgroups to define for this group.
        """,
    )

    validator: Optional[CodeItem] = Field(
        None,
        description="""
A validator to use to validate the testcases of this group.
If not specified, will use the package-level validator.
Useful in cases where the constraints vary across test groups.
""",
    )

    weight: Optional[float] = Field(
        1.0,
        description="""
The weight of this group in the final score. Useful for
problems that have points.
""",
    )


class Generator(CodeItem):
    model_config = ConfigDict(extra='forbid')

    name: str = NameField(description="""The name of the generator.""")


class Solution(CodeItem):
    model_config = ConfigDict(extra='forbid')

    outcome: ExpectedOutcome = Field(
        description="""The expected outcome of this solution."""
    )


class Stress(BaseModel):
    model_config = ConfigDict(extra='forbid')

    name: str = NameField(description='The name of the stress test.')

    generator: GeneratorCall = Field(
        description='Generator pattern to call during stress-test.'
    )

    finder: str = Field(
        description='Finder expression to be used to match against generated tests.'
    )


class LimitModifiers(BaseModel):
    timeMultiplier: Optional[float] = Field(
        None, description='Multiplier for time limit.'
    )
    time: Optional[int] = Field(
        None, description='Value to override time limit with, in milliseconds.'
    )
    memory: Optional[int] = Field(
        None, description='Value to override memory limit with, in MB.'
    )


class Package(BaseModel):
    model_config = ConfigDict(extra='forbid')

    # Name of the problem.
    name: str = NameField(description='The name of the problem.')

    timeLimit: int = Field(description='Time limit of the problem, in milliseconds.')

    memoryLimit: int = Field(description='Memory limit of the problem, in MB.')

    outputLimit: int = Field(
        4 * 1024, description='Output limit of the problem, in KB.'
    )

    modifiers: Dict[str, LimitModifiers] = Field(
        {},
        description="""
    Limit modifiers that can be specified per language.
    """,
    )

    checker: Optional[CodeItem] = Field(
        None, description='The checker for this problem.'
    )

    validator: Optional[CodeItem] = Field(
        None, description='The validator for this problem.'
    )

    generators: List[Generator] = Field([], description='Generators for this problem.')

    solutions: List[Solution] = Field(
        [],
        description="""
All tested solutions for this problem.

The first solution in this list should be the main solution -- the one
that is correct and used as reference -- and should have the `accepted` outcome.
""",
    )

    testcases: List[TestcaseGroup] = Field([], description='Testcases for the problem.')

    stresses: List[Stress] = Field([], description='Stress tests for the problem.')

    statements: List[Statement] = Field([], description='Statements for the problem.')

    # Vars to be re-used across the package.
    #   - It will be passed as --key=value arguments to the validator.
    #   - It will be available as \VAR{key} variables in the rbx statement.
    vars: Dict[str, Primitive] = Field(
        {}, description='Variables to be re-used across the package.'
    )

    @property
    def expanded_vars(self) -> Dict[str, Primitive]:
        return {key: expand_var(value) for key, value in self.vars.items()}

    def timelimit_for_language(self, language: Optional[str]) -> int:
        res = self.timeLimit
        if language is not None and language in self.modifiers:
            modifier = self.modifiers[language]
            if modifier.time is not None:
                res = modifier.time
            if modifier.timeMultiplier is not None:
                res = int(res * float(modifier.timeMultiplier))
        if 'RBX_TIME_MULTIPLIER' in os.environ:
            res = int(res * float(os.environ['RBX_TIME_MULTIPLIER']))
        return res

    def memorylimit_for_language(self, language: Optional[str]) -> int:
        res = self.memoryLimit
        if language is None:
            return res
        if language not in self.modifiers:
            return res
        modifier = self.modifiers[language]
        if modifier.memory is not None:
            return modifier.memory
        return res

    @model_validator(mode='after')
    def check_first_solution_is_main(self):
        if self.solutions:
            if self.solutions[0].outcome != ExpectedOutcome.ACCEPTED:
                raise PydanticCustomError(
                    'MISSING_MAIN_SOLUTION',
                    'The first solution in the package must have the "ACCEPTED" outcome.',
                )
        return self

    @model_validator(mode='after')
    def samples_come_first(self):
        for i, group in enumerate(self.testcases):
            if group.name == 'samples' and i > 0:
                raise PydanticCustomError(
                    'SAMPLES_NOT_FIRST',
                    'The "samples" group must be the first group in the package, but is actually the {i}-th',
                    {'i': i + 1},
                )
        return self



import functools
import importlib
import importlib.resources
import os
import pathlib
import shutil
import subprocess
from typing import Any, Dict, List, Optional

import requests
import typer
from pydantic import BaseModel

from rbx import utils
from rbx.console import console
from rbx.grading.judge.storage import copyfileobj

app = typer.Typer(no_args_is_help=True)

_RESOURCES_PKG = 'rbx.resources'
_CONFIG_FILE_NAME = 'default_config.json'


def format_vars(template: str, **kwargs) -> str:
    res = template
    for key, value in kwargs.items():
        key = key.replace('_', '-')
        res = res.replace(f'%{{{key}}}', value)
    return res


class Artifact(BaseModel):
    filename: Optional[str] = None
    executable: bool = False
    optional: bool = False


class Language(BaseModel):
    template: str
    file: str
    submitFile: Optional[str] = None
    preprocess: Optional[List[str]] = None
    exec: str
    artifacts: Dict[str, Optional[Artifact]] = {}
    submitor: Optional[str] = None

    def get_file(self, basename: str) -> str:
        return format_vars(self.file, problem_code=basename)

    def has_submit_file(self) -> bool:
        return self.submitFile is not None

    def get_submit_file(self, basename: str) -> str:
        if not self.submitFile:
            return self.get_file(basename)
        return format_vars(
            self.submitFile, file=self.get_file(basename), problem_code=basename
        )

    def get_template(self) -> str:
        return get_app_file(pathlib.Path('templates') / self.template).read_text()


SubmitorConfig = Dict[str, Any]
Credentials = Dict[str, Any]


class Config(BaseModel):
    defaultLanguage: str
    languages: Dict[str, Language]
    editor: Optional[str] = None
    submitor: Dict[str, SubmitorConfig]
    credentials: Credentials
    boxEnvironment: str = 'default'

    def get_default_language(self) -> Optional[Language]:
        return self.languages.get(self.defaultLanguage)

    def get_language(self, name: Optional[str] = None) -> Optional[Language]:
        return self.languages.get(name or self.defaultLanguage)


def get_app_path() -> pathlib.Path:
    return utils.get_app_path()


def get_empty_app_persist_path() -> pathlib.Path:
    app_dir = get_app_path() / 'persist'
    shutil.rmtree(str(app_dir), ignore_errors=True)
    app_dir.mkdir(parents=True, exist_ok=True)
    return app_dir


def get_app_file(path: pathlib.Path) -> pathlib.Path:
    file_path = get_app_path() / path
    if file_path.is_file():
        return file_path

    with importlib.resources.as_file(
        importlib.resources.files('rbx') / 'resources' / path  # type: ignore
    ) as file:
        if file.is_file():
            file_path.parent.mkdir(parents=True, exist_ok=True)
            with file.open('rb') as fr:
                with file_path.open('wb') as fw:
                    copyfileobj(fr, fw)
    return file_path


def _download_checker(name: str, save_at: pathlib.Path):
    console.print(f'Downloading checker {name}...')
    r = requests.get(
        f'https://raw.githubusercontent.com/MikeMirzayanov/testlib/master/checkers/{name}'
    )

    if r.ok:
        save_at.parent.mkdir(parents=True, exist_ok=True)
        with save_at.open('wb') as f:
            f.write(r.content)


def _download_testlib(save_at: pathlib.Path):
    console.print('Downloading testlib.h...')
    r = requests.get(
        'https://raw.githubusercontent.com/MikeMirzayanov/testlib/master/testlib.h'
    )

    if r.ok:
        save_at.parent.mkdir(parents=True, exist_ok=True)
        with save_at.open('wb') as f:
            f.write(r.content)
    else:
        console.print('[error]Failed to download testlib.h.[/error]')
        raise typer.Exit(1)


def _download_jngen(save_at: pathlib.Path):
    console.print('Downloading jngen.h...')
    r = requests.get('https://raw.githubusercontent.com/ifsmirnov/jngen/master/jngen.h')

    if r.ok:
        save_at.parent.mkdir(parents=True, exist_ok=True)
        with save_at.open('wb') as f:
            f.write(r.content)
    else:
        console.print('[error]Failed to download jngen.h.[/error]')
        raise typer.Exit(1)


def _download_bits_stdcpp(save_at: pathlib.Path):
    console.print('Downloading bits/stdc++.h...')
    r = requests.get(
        'https://raw.githubusercontent.com/tekfyl/bits-stdc-.h-for-mac/master/stdc%2B%2B.h'
    )

    if r.ok:
        save_at.parent.mkdir(parents=True, exist_ok=True)
        with save_at.open('wb') as f:
            f.write(r.content)
    else:
        console.print('[error]Failed to download bits/stdc++.h.[/error]')
        raise typer.Exit(1)


def get_builtin_checker(name: str) -> pathlib.Path:
    app_file = get_app_file(pathlib.Path('checkers') / name)
    if not app_file.exists():
        _download_checker(name, app_file)
    return app_file


def get_testlib() -> pathlib.Path:
    app_file = get_app_file(pathlib.Path('testlib.h'))
    if not app_file.exists():
        _download_testlib(app_file)
    return app_file


def get_jngen() -> pathlib.Path:
    app_file = get_app_file(pathlib.Path('jngen.h'))
    if not app_file.exists():
        _download_jngen(app_file)
    return app_file


def get_bits_stdcpp() -> pathlib.Path:
    app_file = get_app_file(pathlib.Path('stdc++.h'))
    if not app_file.exists():
        _download_bits_stdcpp(app_file)
    return app_file


def get_default_config_path() -> pathlib.Path:
    with importlib.resources.as_file(
        importlib.resources.files('rbx') / 'resources' / _CONFIG_FILE_NAME
    ) as file:
        return file


def get_default_app_path() -> pathlib.Path:
    return get_default_config_path().parent


def get_default_config() -> Config:
    return Config.model_validate_json(get_default_config_path().read_text())


def get_config_path() -> pathlib.Path:
    return get_app_path() / 'config.json'


def get_editor():
    return get_config().editor or os.environ.get('EDITOR', None)


def open_editor(path: pathlib.Path, *args):
    editor = get_editor()
    if editor is None:
        raise Exception('No editor found. Please set the EDITOR environment variable.')
    subprocess.run([editor, str(path), *[str(arg) for arg in args]])


@functools.cache
def get_config() -> Config:
    config_path = get_config_path()
    if not config_path.is_file():
        utils.create_and_write(config_path, utils.model_json(get_default_config()))
    return Config.model_validate_json(config_path.read_text())


def save_config(cfg: Config):
    cfg_path = get_config_path()
    cfg_path.write_text(utils.model_json(cfg))
    get_config.cache_clear()


@app.command()
def path():
    """
    Show the absolute path of the config file.
    """
    get_config()  # Ensure config is created.
    console.print(get_config_path())


@app.command('list, ls')
def list():
    """
    Pretty print the config file.
    """
    console.print_json(utils.model_json(get_config()))


@app.command()
def reset():
    """
    Reset the config file to the default one.
    """
    if not typer.confirm('Do you really want to reset your config to the default one?'):
        return
    cfg_path = get_config_path()
    cfg_path.unlink(missing_ok=True)
    get_config()  # Reset the config.


@app.command('edit, e')
def edit():
    """
    Open the config in an editor.
    """
    open_editor(get_config_path())



import pathlib
from typing import Any

import mechanize

from rbx.console import console
from rbx.schema import Problem
from rbx.submitors.submitor import Submitor

_SUBMITOR_KEY = 'codeforces'


class Session:
    def __init__(self):
        self.br = mechanize.Browser()
        self.br.set_handle_robots(False)
        self.br.addheaders = [  # type: ignore
            (
                'User-agent',
                'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.1) Gecko/2008071615 Fedora/3.0.1-1.fc9 Firefox/3.0.1',
            )
        ]

    def select_form_by_id(self, id, forms):
        for form in forms:
            if form.attrs.get('id') == id:
                return form
        return None

    def select_form_by_class(self, cl, forms):
        for form in forms:
            if form.attrs.get('class') == cl:
                return form
        return None

    def login(self, handle, password):
        response = self.br.open('http://codeforces.com/enter')
        if response is None:
            raise Exception('Response was not received')
        if not response.geturl().endswith('enter'):
            return
        form = self.select_form_by_id('enterForm', self.br.forms())
        if form is None:
            raise Exception('Login form was not found')

        form['handleOrEmail'] = handle
        form['password'] = password
        self.br.form = form
        response = self.br.submit()
        if response is None:
            raise Exception('Response was not received')
        if response.geturl().endswith('enter'):
            raise Exception(
                'Login attempt was not successful. Check your credentials or your internet connection'
            )

    def submit_from_contest(self, url: str, file: pathlib.Path, typeid: int):
        filename = str(file.resolve())
        self.br.open(url)
        form = self.select_form_by_class('submit-form', self.br.forms())
        if form is None:
            raise Exception('You are not logged in or problem does not exist')

        form['programTypeId'] = [str(typeid)]
        form.add_file(file.open(), 'plain/text', filename)
        self.br.form = form
        self.br.submit()

    def submit(self, url: str, file: pathlib.Path, typeid: int):
        filename = str(file.resolve())
        response = self.br.open(url)
        if response is None:
            raise Exception('Response was not received')
        if response.geturl().endswith('attachments'):
            self.submit_from_contest(url.replace('/problem/', '/submit/'), file, typeid)
            return
        form = self.select_form_by_class('submitForm', self.br.forms())
        if form is None:
            raise Exception('You are not logged in or problem does not exist')

        form['programTypeId'] = [str(typeid)]
        form.add_file(file.open(), 'plain/text', filename)
        self.br.form = form
        self.br.submit()


class CodeforcesSubmitor(Submitor):
    def key(self) -> str:
        return _SUBMITOR_KEY

    def should_handle(self, problem: Problem) -> bool:
        return 'codeforces.com/' in problem.url

    def submit(
        self,
        file: pathlib.Path,
        problem: Problem,
        submitor_config: Any,
        credentials: Any,
    ) -> bool:
        typeid = submitor_config['typeid']
        handle = credentials['handle']
        password = credentials['password']

        session = Session()
        console.print(
            f'Trying to login with handle [item]{handle}[/item] in Codeforces.'
        )
        try:
            session.login(handle, password)
        except Exception as e:
            console.print(f'[error]Failed to log in: {str(e)}[/error]')
            return False
        console.print('[italic]Possibly[/italic] logged in!')
        try:
            session.submit(problem.url, file, typeid)
        except Exception as e:
            console.print(f'[error]Failed to submit: {str(e)}[/error]')
            return False
        return True



import functools
import os
import pathlib
import shlex
import shutil
import subprocess
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel
from rich.text import Text

from rbx import utils
from rbx.config import get_bits_stdcpp, get_jngen, get_testlib
from rbx.console import console
from rbx.grading.judge.sandbox import SandboxBase, SandboxParams
from rbx.grading.judge.storage import copyfileobj

MAX_STDOUT_LEN = 1024 * 1024 * 128  # 128 MB


class Outcome(Enum):
    ACCEPTED = 'accepted'
    WRONG_ANSWER = 'wrong-answer'
    JUDGE_FAILED = 'judge-failed'
    RUNTIME_ERROR = 'runtime-error'
    TIME_LIMIT_EXCEEDED = 'time-limit-exceeded'
    MEMORY_LIMIT_EXCEEDED = 'memory-limit-exceeded'
    OUTPUT_LIMIT_EXCEEDED = 'output-limit-exceeded'
    INTERNAL_ERROR = 'internal-error'


class DigestHolder(BaseModel):
    value: Optional[str] = None


class GradingLogsHolder(BaseModel):
    run: Optional['RunLog'] = None
    cached: bool = False


class DigestOrSource(BaseModel):
    # Source path relative to the FS.
    src: Optional[pathlib.Path] = None
    # Digest if we should get file from storage.
    digest: Optional[DigestHolder] = None

    @staticmethod
    def create(data: Union[pathlib.Path, DigestHolder, str]) -> 'DigestOrSource':
        if isinstance(data, str):
            return DigestOrSource(digest=DigestHolder(value=data))
        if isinstance(data, DigestHolder):
            return DigestOrSource(digest=data)
        return DigestOrSource(src=data)

    def expand(self) -> Dict[str, Any]:
        res = {}
        if self.src is not None:
            res['src'] = self.src
        if self.digest is not None:
            res['digest'] = self.digest
        return res


class DigestOrDest(BaseModel):
    # Destination path relative to the FS.
    dest: Optional[pathlib.Path] = None
    # Digest if we should get file from storage.
    digest: Optional[DigestHolder] = None

    @staticmethod
    def create(data: Union[pathlib.Path, DigestHolder, str]) -> 'DigestOrDest':
        if isinstance(data, str):
            return DigestOrDest(digest=DigestHolder(value=data))
        if isinstance(data, DigestHolder):
            return DigestOrDest(digest=data)
        return DigestOrDest(dest=data)

    def expand(self) -> Dict[str, Any]:
        res = {}
        if self.dest is not None:
            res['dest'] = self.dest
        if self.digest is not None:
            res['digest'] = self.digest
        return res


class GradingFileInput(BaseModel):
    # Destination path relative to the sandboox.
    dest: pathlib.Path
    # Source path relative to the FS.
    src: Optional[pathlib.Path] = None
    # Digest if we should get file from storage.
    digest: Optional[DigestHolder] = None
    # Whether the destination file should be marked as an executable.
    executable: bool = False


class GradingFileOutput(BaseModel):
    # Source path relative to the sandbox.
    src: pathlib.Path
    # Destination path relative to the FS.
    dest: Optional[pathlib.Path] = None
    # Digest if we should put file in storage.
    digest: Optional[DigestHolder] = None
    # Whether the destination file should be marked as an executable.
    executable: bool = False
    # Whether the file is optional or not.
    optional: bool = False
    # Whether to cap its size
    maxlen: Optional[int] = None
    # Whether the file is just an intermediate file that should not be tracked.
    intermediate: bool = False
    # Whether to track file through its hash (disable for optimization).
    hash: bool = True


class GradingArtifacts(BaseModel):
    # Root directory for the produced artifacts.
    root: pathlib.Path = pathlib.PosixPath('.')
    # List of input files to copy to the sandbox.
    inputs: List[GradingFileInput] = []
    # List of output files to copy from the sandbox.
    outputs: List[GradingFileOutput] = []
    # Capture certain logs of the execution.
    logs: Optional[GradingLogsHolder] = None


class TestcaseIO(BaseModel):
    index: int
    input: Optional[pathlib.Path] = None
    output: Optional[pathlib.Path] = None


class RunLogMetadata(BaseModel):
    language: Optional[str] = None


class RunLog(BaseModel):
    exitcode: int = 0
    exitstatus: str = SandboxBase.EXIT_SANDBOX_ERROR
    time: Optional[float] = 0.0
    memory: Optional[int] = 0
    metadata: Optional[RunLogMetadata] = None

    def get_run_language(self) -> Optional[str]:
        if self.metadata is None:
            return None
        return self.metadata.language

    def get_summary(self) -> str:
        if self.exitcode == 0:
            return 'OK'
        time = self.time or 0.0
        memory = self.memory or 0
        return f'FAILED with exit code {self.exitcode} and sandbox status {self.exitstatus} (time: {time}s, memory: {memory//(1024*1024)}MB)'


class PreprocessLog(RunLog):
    cmd: List[str]
    log: str


class TestcaseLog(RunLog):
    stdout_absolute_path: Optional[pathlib.Path] = None
    stderr_absolute_path: Optional[pathlib.Path] = None
    log_absolute_path: Optional[pathlib.Path] = None


class CheckerResult(BaseModel):
    outcome: Outcome
    message: str = ''
    no_tle_outcome: Optional[Outcome] = None


class Evaluation(BaseModel):
    result: CheckerResult
    testcase: TestcaseIO
    log: TestcaseLog


def _process_input_artifacts(artifacts: GradingArtifacts, sandbox: SandboxBase):
    for input_artifact in artifacts.inputs:
        if input_artifact.digest is not None:
            assert input_artifact.digest.value is not None
            sandbox.create_file_from_storage(
                input_artifact.dest,
                input_artifact.digest.value,
                override=True,
                executable=input_artifact.executable,
                try_symlink=True,
            )
            continue
        assert input_artifact.src is not None
        sandbox.create_file_from_other_file(
            input_artifact.dest,
            artifacts.root / input_artifact.src,
            executable=input_artifact.executable,
            override=True,
            try_symlink=True,
        )


def _process_output_artifacts(
    artifacts: GradingArtifacts, sandbox: SandboxBase
) -> bool:
    for output_artifact in artifacts.outputs:
        if output_artifact.hash and output_artifact.digest is None:
            output_artifact.digest = DigestHolder()
        if not sandbox.file_exists(output_artifact.src):
            if output_artifact.optional:
                continue
            console.print(
                f'[error]Output artifact [item]{output_artifact.src}[/item] does not exist.[/error]'
            )
            return False

        if output_artifact.digest is not None:
            output_artifact.digest.value = sandbox.get_file_to_storage(
                output_artifact.src, trunc_len=output_artifact.maxlen
            )
        if output_artifact.dest is None:
            continue
        dst: pathlib.Path = artifacts.root / output_artifact.dest
        # Ensure dst directory exists.
        dst.parent.mkdir(parents=True, exist_ok=True)
        with dst.open('wb') as f:
            with sandbox.get_file(output_artifact.src) as sb_f:
                copyfileobj(
                    sb_f,
                    f,
                    maxlen=output_artifact.maxlen,
                )
        if output_artifact.executable:
            dst.chmod(0o755)
    return True


def testlib_grading_input() -> GradingFileInput:
    return GradingFileInput(src=get_testlib(), dest=pathlib.Path('testlib.h'))


def jngen_grading_input() -> GradingFileInput:
    return GradingFileInput(src=get_jngen(), dest=pathlib.Path('jngen.h'))


def _expand_part(part: str, sandbox: SandboxBase) -> List[str]:
    part = part.strip()
    if part.startswith('@glob:'):
        return [shlex.quote(str(path)) for path in sandbox.glob(part[6:])]
    return [part]


def _split_and_expand(command: str, sandbox: SandboxBase) -> List[str]:
    res = []
    parts = shlex.split(command.format(memory=sandbox.params.address_space))
    for part in parts:
        res.extend(_expand_part(part, sandbox))
    return res


def _is_c_command(exe_command: str) -> bool:
    return exe_command.endswith('gcc') or exe_command.endswith('clang')


def _is_cpp_command(exe_command: str) -> bool:
    return exe_command.endswith('g++') or exe_command.endswith('clang++')


@functools.cache
def _complain_about_clang() -> None:
    console.print(
        '[warning]Notice your C++ files are being compiled with [item]clang[/item] instead of [item]g++[/item].[/warning]'
    )
    console.print('[warning]This may lead to unexpected behavior.[/warning]')
    console.print('[warning]Consider using [item]g++[/item] instead.[/warning]')
    console.print(
        '[warning]See [item]https://rsalesc.github.io/rbx/cpp-on-macos[/item] for instructions on how to use [item]g++[/item] on MacOS.'
    )


def _maybe_get_bits_stdcpp_for_clang(command: str) -> Optional[GradingFileInput]:
    cmds = shlex.split(command)
    if not cmds:
        return None
    exe = cmds[0]

    if not _is_cpp_command(exe):
        return None

    output = subprocess.run([exe, '-v'], capture_output=True)
    if output.returncode != 0:
        console.print('[error]Failed to get g++/clang compiler version.[/error]')
        return None
    lines = output.stderr.decode().splitlines()
    if not lines:
        return None
    # Check the first line for `clang`.
    if 'clang' not in lines[0]:
        return None

    _complain_about_clang()
    bits = get_bits_stdcpp()
    return GradingFileInput(src=bits, dest=pathlib.Path('bits/stdc++.h'))


def _maybe_get_bits_stdcpp_for_commands(
    commands: List[str],
) -> Optional[GradingFileInput]:
    for command in commands:
        res = _maybe_get_bits_stdcpp_for_clang(command)
        if res is not None:
            return res
    return None


@functools.cache
def _try_following_alias_for_exe(exe: str) -> Optional[str]:
    if _is_c_command(exe) and os.environ.get('RBX_C_PATH'):
        return os.environ['RBX_C_PATH']
    if _is_cpp_command(exe) and os.environ.get('RBX_CXX_PATH'):
        return os.environ['RBX_CXX_PATH']
    output = subprocess.run(
        f'which {exe}', shell=True, executable=shutil.which('bash'), capture_output=True
    )
    if output.returncode != 0:
        return None
    return output.stdout.decode().strip()


def _try_following_alias_for_command(command: str) -> str:
    cmds = shlex.split(command)
    if not cmds:
        return command
    exe = cmds[0]
    new_exe = _try_following_alias_for_exe(exe)
    if new_exe is None:
        return command
    return shlex.join([new_exe, *cmds[1:]])


def _try_following_alias_for_commands(commands: List[str]) -> List[str]:
    res: List[str] = []
    for command in commands:
        res.append(_try_following_alias_for_command(command))
    return res


def compile(
    commands: List[str],
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
) -> bool:
    commands = _try_following_alias_for_commands(commands)
    bits_artifact = _maybe_get_bits_stdcpp_for_commands(commands)
    if bits_artifact is not None:
        _process_input_artifacts(GradingArtifacts(inputs=[bits_artifact]), sandbox)
    _process_input_artifacts(artifacts, sandbox)

    if not commands:
        # Code does not need preprocessing of any kind.
        return True

    logs: List[PreprocessLog] = []
    sandbox.set_params(params)

    for i, command in enumerate(commands):
        cmd = _split_and_expand(command, sandbox)
        stdout_file = pathlib.PosixPath(f'compile-{i}.stdout')
        stderr_file = pathlib.PosixPath(f'compile-{i}.stderr')
        sandbox.params.set_stdall(stdout=stdout_file, stderr=stderr_file)

        if bits_artifact is not None and _is_cpp_command(cmd[0]):
            # Include from sandbox directory to import bits/stdc++.h.
            cmd.append('-I.')

        if not sandbox.execute_without_std(cmd):
            console.print(
                '[error]Sandbox crashed while processing command:[/error]',
                utils.highlight_json_obj(cmd),
                '[error]and logs[/error]',
                sandbox.debug_message(),
            )
            return False

        std_outputs = [
            sandbox.get_file_to_string(stderr_file, maxlen=None)
            if sandbox.file_exists(stderr_file)
            else '<No stderr produced by command>',
            sandbox.get_file_to_string(stdout_file, maxlen=None)
            if sandbox.file_exists(stdout_file)
            else '<No stdout produced by command>',
        ]

        log = PreprocessLog(
            cmd=cmd,
            exitcode=sandbox.get_exit_code(),
            exitstatus=sandbox.get_exit_status(),
            time=sandbox.get_execution_time(),
            memory=sandbox.get_memory_used(),
            log='\n'.join(std_outputs),
        )
        logs.append(log)

        if log.exitcode != 0:
            break

    if logs and logs[-1].exitcode != 0:
        console.print(
            '[error]FAILED[/error] Preprocessing failed with command',
            utils.highlight_json_obj(logs[-1].cmd),
        )
        console.print(f'[error]Summary:[/error] {logs[-1].get_summary()}')
        console.print(Text.from_ansi(logs[-1].log), style='default')
        return False

    return _process_output_artifacts(artifacts, sandbox)


def run(
    command: str,
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
    metadata: Optional[RunLogMetadata] = None,
) -> Optional[RunLog]:
    _process_input_artifacts(artifacts, sandbox)
    cmd = _split_and_expand(command, sandbox)
    sandbox.set_params(params)

    if not sandbox.execute_without_std(cmd):
        console.print(
            '[error]Sandbox crashed while processing command:[/error]',
            utils.highlight_json_obj(cmd),
            '[error]and logs[/error]',
            sandbox.debug_message(),
        )
        return None

    if not _process_output_artifacts(artifacts, sandbox):
        return None

    execution_time = sandbox.get_execution_time()
    if execution_time is not None and (
        sandbox.get_exit_status() == SandboxBase.EXIT_TIMEOUT
        or sandbox.get_exit_status() == SandboxBase.EXIT_TIMEOUT_WALL
    ):
        execution_time = max(execution_time, (params.timeout or 0.0) / 1000)

    run_log = RunLog(
        exitcode=sandbox.get_exit_code(),
        exitstatus=sandbox.get_exit_status(),
        time=sandbox.get_execution_time(),
        memory=sandbox.get_memory_used(),
        metadata=metadata,
    )
    if artifacts.logs is not None:
        artifacts.logs.run = run_log.model_copy()
    return run_log


def _normalize_checked_words(s: str) -> Tuple[str, ...]:
    return tuple(s.split())


def _wcmp_check(expected: str, output: str) -> Outcome:
    if _normalize_checked_words(expected) == _normalize_checked_words(output):
        return Outcome.ACCEPTED

    return Outcome.WRONG_ANSWER


def get_checker_sandbox_params() -> SandboxParams:
    params = SandboxParams(
        max_processes=None,
        preserve_env=True,
    )
    params.add_mapped_directory(pathlib.Path('/usr'))
    params.add_mapped_directory(pathlib.Path('/etc'))
    return params


def _check(
    sandbox: SandboxBase,
    testcase: TestcaseIO,
    output_path: pathlib.Path,
    should_use_python_checker: bool = False,
) -> CheckerResult:
    if testcase.output is None:
        # No output to compare.
        return CheckerResult(outcome=Outcome.ACCEPTED)

    if should_use_python_checker:
        # Use default wcmp checker.
        expected = testcase.output.read_text()
        output = output_path.read_text()

        return CheckerResult(outcome=_wcmp_check(expected, output))

    sandbox.params.set_stdall(
        stdin=None,
        stdout=pathlib.PosixPath('stdout.txt'),
        stderr=pathlib.PosixPath('stderr.txt'),
    )

    sandbox.create_file_from_string(
        pathlib.PosixPath('expected.txt'), testcase.output.read_text(), override=True
    )
    sandbox.create_file_from_string(
        pathlib.PosixPath('output.txt'), output_path.read_text(), override=True
    )
    sandbox.create_file_from_string(
        pathlib.PosixPath('input.txt'),
        testcase.input.read_text() if testcase.input is not None else '',
        override=True,
    )

    if not sandbox.execute_without_std(
        ['./checker', 'input.txt', 'output.txt', 'expected.txt'],
    ):
        console.print(
            '[error]Sandbox crashed while running checker.[/error]',
            '[error]and logs[/error]',
            sandbox.debug_message(),
        )
        return CheckerResult(outcome=Outcome.INTERNAL_ERROR)

    checker_stderr = sandbox.get_file_to_string(
        pathlib.PosixPath('stderr.txt'), maxlen=None
    )
    if sandbox.get_exit_code() in [1, 2]:
        return CheckerResult(outcome=Outcome.WRONG_ANSWER, message=checker_stderr)
    if sandbox.get_exit_code() == 3:
        return CheckerResult(outcome=Outcome.JUDGE_FAILED, message=checker_stderr)
    return CheckerResult(outcome=Outcome.ACCEPTED, message=checker_stderr)


# Always assume a `checker` executable in the sandbox if should use checker.
def evaluate(
    sandbox: SandboxBase,
    testcase: TestcaseIO,
    log: Optional[TestcaseLog],
    artifacts: GradingArtifacts,
    should_use_python_checker: bool = False,
) -> Evaluation:
    if log is None:
        return Evaluation(
            testcase=testcase,
            log=TestcaseLog(),
            result=CheckerResult(outcome=Outcome.INTERNAL_ERROR),
        )
    if log.exitstatus != SandboxBase.EXIT_OK:
        return Evaluation(
            testcase=testcase,
            log=log,
            result=CheckerResult(outcome=Outcome.RUNTIME_ERROR),
        )

    if not testcase.output:
        # No output to compare.
        return Evaluation(
            testcase=testcase, log=log, result=CheckerResult(outcome=Outcome.ACCEPTED)
        )

    _process_input_artifacts(artifacts, sandbox)
    if log.stdout_absolute_path is None:
        return Evaluation(
            testcase=testcase,
            log=log,
            result=CheckerResult(outcome=Outcome.INTERNAL_ERROR, message='No output'),
        )

    checker_result = _check(
        sandbox,
        testcase,
        log.stdout_absolute_path,
        should_use_python_checker=should_use_python_checker,
    )
    return Evaluation(
        testcase=testcase,
        log=log,
        result=checker_result,
    )



import os
import pathlib
import shutil
from collections.abc import Iterator

import pytest

from rbx import testing_utils
from rbx.box import package


@pytest.fixture
def pkg_cleandir(cleandir: pathlib.Path) -> Iterator[pathlib.Path]:
    old_temp_dir = package.TEMP_DIR
    package.TEMP_DIR = cleandir

    pkgdir = cleandir / 'pkg'
    pkgdir.mkdir(exist_ok=True, parents=True)
    lwd = pathlib.Path.cwd()
    os.chdir(str(pkgdir))
    try:
        yield pkgdir.absolute()
    finally:
        os.chdir(str(lwd))
        package.TEMP_DIR = old_temp_dir


@pytest.fixture
def pkg_from_testdata(
    request, testdata_path: pathlib.Path, pkg_cleandir: pathlib.Path
) -> Iterator[pathlib.Path]:
    marker = request.node.get_closest_marker('test_pkg')
    if marker is None:
        raise ValueError('test_pkg marker not found')
    testdata = testdata_path / marker.args[0]
    shutil.copytree(str(testdata), str(pkg_cleandir), dirs_exist_ok=True)
    yield pkg_cleandir


@pytest.fixture(autouse=True)
def clear_cache():
    testing_utils.clear_all_functools_cache()



import pathlib
from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from rbx.box.schema import NameField, Primitive, expand_var
from rbx.box.statements.schema import (
    ConversionStep,
    Joiner,
    StatementType,
)


def ShortNameField(**kwargs):
    return Field(pattern=r'^[A-Z]+[0-9]*$', min_length=1, max_length=4, **kwargs)


class ProblemStatementOverride(BaseModel):
    model_config = ConfigDict(extra='forbid')

    configure: List[ConversionStep] = Field(
        [],
        discriminator='type',
        description="""
Configure how certain conversion steps should happen when applied to the statement file.

Different from the `steps` field, this does not force the steps to happen, but rather only
configure them in case they are applied.
""",
    )


class ContestStatement(BaseModel):
    model_config = ConfigDict(extra='forbid')

    language: str = Field('en', description='Language code for this statement.')

    title: str = Field(description='Title of the contest in this language.')

    location: Optional[str] = Field(
        default=None, description='Location of the contest in this language.'
    )

    date: Optional[str] = Field(
        default=None, description='Date of the contest in this language.'
    )

    path: pathlib.Path = Field(description='Path to the input statement file.')

    type: StatementType = Field(description='Type of the input statement file.')

    joiner: Optional[Joiner] = Field(
        None,
        description="""
Joiner to be used to build the statement.
                           
This determines how problem statements will be joined into a single contest statement.""",
    )

    steps: List[ConversionStep] = Field(
        default=[],
        discriminator='type',
        description="""
Describes a sequence of conversion steps that should be applied to the statement file
of this contest.

Usually, it is not necessary to specify these, as they can be inferred from the
input statement type and the output statement type, but you can use this to force
certain conversion steps to happen.
""",
    )

    configure: List[ConversionStep] = Field(
        default=[],
        discriminator='type',
        description="""
Configure how certain conversion steps should happen when applied to the statement file of
this contest.

Different from the `steps` field, this does not force the steps to happen, but rather only
configure them in case they are applied.
""",
    )

    assets: List[str] = Field(
        default=[],
        description="""
Assets relative to the contest directory that should be included while building
the statement. Files will be included in the same folder as the statement file.
Can be glob pattern as well, such as `imgs/*.png`.
""",
    )

    override: Optional[ProblemStatementOverride] = Field(
        default=None, description='Override configuration for problem statements.'
    )

    # Vars to be re-used in the statement.
    #   - It will be available as \VAR{vars} variable in the contest-level box statement.
    vars: Dict[str, Primitive] = Field(
        {}, description='Variables to be re-used across the package.'
    )

    @property
    def expanded_vars(self) -> Dict[str, Primitive]:
        return {key: expand_var(value) for key, value in self.vars.items()}


class ContestProblem(BaseModel):
    short_name: str = ShortNameField(
        description="""
Short name of the problem. Usually, just an uppercase letter,
but can be a sequence of uppercase letters followed by a number."""
    )
    path: Optional[pathlib.Path] = Field(
        default=None,
        description="""
Path to the problem relative to the contest package directory.
If not specified, will expect the problem to be in ./{short_name}/ folder.""",
    )

    color: Optional[str] = Field(
        default=None,
        description="""Hex-based color that represents this problem in the contest.""",
        pattern=r'^[A-Za-z0-9]+$',
        max_length=6,
    )

    def get_path(self) -> pathlib.Path:
        return self.path or pathlib.Path(self.short_name)


class Contest(BaseModel):
    model_config = ConfigDict(extra='forbid')

    name: str = NameField(description='Name of this contest.')

    problems: List[ContestProblem] = Field(
        default=[], description='List of problems in this contest.'
    )

    statements: List[ContestStatement] = Field(
        default=None,
        description='Configure statements in this contest, per language.',
    )

    # Vars to be re-used in the statements.
    #   - It will be available as \VAR{vars} variable in the contest-level box statement.
    vars: Dict[str, Primitive] = Field(
        {}, description='Variables to be re-used across the package.'
    )

    @property
    def expanded_vars(self) -> Dict[str, Primitive]:
        return {key: expand_var(value) for key, value in self.vars.items()}



import pathlib
from typing import Optional

from rbx import annotations, console, metadata
from rbx.config import get_config, open_editor


def main(problem: str, language: Optional[annotations.LanguageWithDefault] = None):
    lang = get_config().get_language(language)
    if lang is None:
        console.console.print(
            f'[error]Language {language or get_config().defaultLanguage} not found in config. Please check your configuration.[/error]'
        )
        return

    dumped_problem = metadata.find_problem_by_anything(problem)
    if not dumped_problem:
        console.console.print(
            f'[error]Problem with identifier {problem} not found.[/error]'
        )
        return

    filename = lang.get_file(dumped_problem.code)
    open_editor(pathlib.Path(filename))



import dataclasses
import pathlib
import re
import shutil
import typing
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple

import typer

from rbx import console
from rbx.box.schema import Package, Primitive, Testcase
from rbx.box.statements.latex import (
    MAX_PDFLATEX_RUNS,
    Latex,
    decode_latex_output,
    should_rerun,
)
from rbx.box.statements.latex_jinja import (
    JinjaDictWrapper,
    render_latex_template,
    render_latex_template_blocks,
)
from rbx.box.statements.schema import (
    ConversionStep,
    ConversionType,
    JinjaTeX,
    Statement,
    StatementType,
    TexToPDF,
    rbxToTeX,
)


@dataclasses.dataclass
class StatementCodeLanguage:
    id: str
    name: str
    command: str


@dataclasses.dataclass
class StatementBuilderContext:
    languages: List[StatementCodeLanguage]
    params: ConversionStep
    root: pathlib.Path
    editorial: bool
    vars: Optional[Dict[str, Primitive]] = None

    def build_jinja_kwargs(self) -> Dict[str, Any]:
        res = {
            'languages': self.languages,
            'keyed_languages': {lang.id: lang for lang in self.languages},
            'is_editorial': self.editorial,
        }
        if self.vars is not None:
            res['vars'] = self.vars
        return res


class StatementBuilderItem(ABC):
    @abstractmethod
    def build_jinja_kwargs(self) -> Dict[str, Any]:
        pass


class StatementSample(Testcase):
    explanation: Optional[str] = None


@dataclasses.dataclass
class StatementBuilderProblem(StatementBuilderItem):
    package: Package
    statement: Statement
    samples: List[Testcase] = dataclasses.field(default_factory=list)
    short_name: Optional[str] = None

    # Will only be filled by contests.
    io_path: Optional[pathlib.Path] = None

    def build_inner_jinja_kwargs(self) -> Dict[str, Any]:
        kwargs = {
            'package': self.package,
            'statement': self.statement,
            'samples': self.samples,
            'vars': JinjaDictWrapper(self.package.expanded_vars, key='vars'),
            'title': self.statement.title or self.package.name,
        }
        if self.short_name is not None:
            kwargs['short_name'] = self.short_name
        if self.io_path is not None:
            kwargs['path'] = self.io_path
        return kwargs

    def build_jinja_kwargs(self) -> Dict[str, Any]:
        inner = self.build_inner_jinja_kwargs()
        return {
            'problem': inner,
        }


@dataclasses.dataclass
class StatementBuilderContest(StatementBuilderItem):
    title: str
    location: Optional[str] = None
    date: Optional[str] = None
    problems: List[StatementBuilderProblem] = dataclasses.field(default_factory=list)

    def build_inner_jinja_kwargs(self) -> Dict[str, Any]:
        res = {'title': self.title}
        if self.location:
            res['location'] = self.location
        if self.date:
            res['date'] = self.date
        return res

    def build_jinja_kwargs(self) -> Dict[str, Any]:
        res = {
            'contest': self.build_inner_jinja_kwargs(),
            'problems': [
                problem.build_inner_jinja_kwargs() for problem in self.problems
            ],
        }
        return res


@dataclasses.dataclass
class StatementBlocks:
    blocks: Dict[str, str] = dataclasses.field(default_factory=dict)
    explanations: Dict[int, str] = dataclasses.field(default_factory=dict)


def prepare_assets(
    assets: List[Tuple[pathlib.Path, pathlib.Path]],
    dest_dir: pathlib.Path,
):
    dest_dir.mkdir(parents=True, exist_ok=True)

    for asset_in, asset_out in assets:
        if not asset_in.is_file():
            console.console.print(
                f'[error]Asset [item]{asset_in}[/item] does not exist in your package.[/error]'
            )
            raise typer.Exit(1)

        # dest_path = dest_dir / asset.resolve().relative_to(statement_dir)
        dest_path = dest_dir / asset_out
        dest_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(str(asset_in), str(dest_path))


def render_jinja(root: pathlib.Path, content: bytes, **kwargs) -> bytes:
    temp_file = '__input__.tex'
    temp_path = root / temp_file
    temp_path.write_bytes(content)

    result: str = render_latex_template(
        str(root),
        temp_file,
        kwargs,
    )
    return result.encode()


def render_jinja_blocks(
    root: pathlib.Path, content: bytes, **kwargs
) -> StatementBlocks:
    temp_file = '__input__.tex'
    temp_path = root / temp_file
    temp_path.write_bytes(content)

    result: Dict[str, str] = render_latex_template_blocks(
        str(root),
        temp_file,
        kwargs,
    )

    pattern = re.compile(r'explanation_(\d+)')
    explanation_keys = []
    for key in result:
        if match := pattern.match(key):
            explanation_keys.append((key, int(match.group(1))))

    explanations = {value: result[key] for key, value in explanation_keys}
    return StatementBlocks(blocks=result, explanations=explanations)


class StatementBuilder(ABC):
    @abstractmethod
    def name(self) -> ConversionType:
        pass

    @abstractmethod
    def default_params(self) -> ConversionStep:
        pass

    @abstractmethod
    def input_type(self) -> StatementType:
        pass

    @abstractmethod
    def output_type(self) -> StatementType:
        pass

    def handles_contest(self) -> bool:
        return True

    def handles_problem(self) -> bool:
        return True

    def inject_assets(
        self, root: pathlib.Path, params: ConversionStep
    ) -> List[Tuple[pathlib.Path, pathlib.Path]]:
        return []

    @abstractmethod
    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        pass


class JinjaTeXBuilder(StatementBuilder):
    def name(self) -> ConversionType:
        return ConversionType.JinjaTeX

    def default_params(self) -> ConversionStep:
        return JinjaTeX(type=ConversionType.JinjaTeX)

    def input_type(self) -> StatementType:
        return StatementType.JinjaTeX

    def output_type(self) -> StatementType:
        return StatementType.TeX

    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        return render_jinja(
            context.root,
            input,
            **context.build_jinja_kwargs(),
            **item.build_jinja_kwargs(),
        )


class rbxTeXBuilder(StatementBuilder):
    def name(self) -> ConversionType:
        return ConversionType.rbxToTex

    def default_params(self) -> ConversionStep:
        return rbxToTeX(type=ConversionType.rbxToTex)

    def input_type(self) -> StatementType:
        return StatementType.rbxTeX

    def output_type(self) -> StatementType:
        return StatementType.TeX

    def handles_contest(self) -> bool:
        # This builder cannot build contest statements.
        return False

    def inject_assets(
        self, root: pathlib.Path, params: ConversionStep
    ) -> List[Tuple[pathlib.Path, pathlib.Path]]:
        params = typing.cast(rbxToTeX, params)
        if not params.template:
            return []
        return [((root / params.template).resolve(), params.template)]

    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        params = typing.cast(rbxToTeX, context.params)
        assert params.template is not None
        problem = typing.cast(StatementBuilderProblem, item)

        statement_blocks = render_jinja_blocks(
            context.root, input, **problem.build_inner_jinja_kwargs()
        )
        blocks = statement_blocks.blocks

        # Remove editorial block when not editorial.
        if not context.editorial and 'editorial' in blocks:
            del blocks['editorial']

        problem_kwargs = problem.build_jinja_kwargs()
        problem_kwargs['problem']['blocks'] = blocks
        if statement_blocks.explanations is not None:
            problem_kwargs['problem']['samples'] = [
                StatementSample(
                    **typing.cast(Testcase, sample).model_dump(),
                    explanation=statement_blocks.explanations.get(i),
                )
                for i, sample in enumerate(problem_kwargs['problem']['samples'])
            ]

        return render_jinja(
            context.root,
            f'%- extends "{params.template}"'.encode(),
            **context.build_jinja_kwargs(),
            **problem_kwargs,
        )


class TeX2PDFBuilder(StatementBuilder):
    def name(self) -> ConversionType:
        return ConversionType.TexToPDF

    def default_params(self) -> ConversionStep:
        return TexToPDF(type=ConversionType.TexToPDF)

    def input_type(self) -> StatementType:
        return StatementType.TeX

    def output_type(self) -> StatementType:
        return StatementType.PDF

    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        latex = Latex(input.decode())
        latex_result = latex.build_pdf(context.root)
        pdf = latex_result.pdf
        logs = decode_latex_output(latex_result.result.stdout)
        runs = 1

        while pdf is not None and should_rerun(logs) and runs < MAX_PDFLATEX_RUNS:
            console.console.print(
                'Re-running pdfLaTeX to get cross-references right...'
            )
            latex_result = latex.build_pdf(context.root)
            pdf = latex_result.pdf
            logs = decode_latex_output(latex_result.result.stdout)
            runs += 1

        if pdf is None:
            console.console.print(f'{logs}')
            console.console.print('[error]PdfLaTeX compilation failed.[/error]')
            raise typer.Exit(1)

        if verbose:
            console.console.print(f'{logs}')

        return pdf


BUILDER_LIST: List[StatementBuilder] = [
    TeX2PDFBuilder(),
    JinjaTeXBuilder(),
    rbxTeXBuilder(),
]
PROBLEM_BUILDER_LIST = [
    builder for builder in BUILDER_LIST if builder.handles_problem()
]
CONTEST_BUILDER_LIST = [
    builder for builder in BUILDER_LIST if builder.handles_contest()
]



import pathlib
import shutil
import subprocess
from typing import Annotated, Optional

import rich.prompt
import typer

from rbx import annotations, console, utils
from rbx.box import creation, presets
from rbx.box.contest import contest_utils, statements
from rbx.box.contest.contest_package import (
    find_contest,
    find_contest_package_or_die,
    find_contest_yaml,
    save_contest,
    within_contest,
)
from rbx.box.contest.schema import ContestProblem
from rbx.box.packaging import contest_main as packaging
from rbx.box.presets.fetch import get_preset_fetch_info
from rbx.box.schema import Package
from rbx.config import open_editor

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)
app.add_typer(
    statements.app,
    name='statements, st',
    cls=annotations.AliasGroup,
    help='Manage contest-level statements.',
)
app.add_typer(
    packaging.app,
    name='package, pkg',
    cls=annotations.AliasGroup,
    help='Build contest-level packages.',
)


@app.command('create, c', help='Create a new contest package.')
def create(
    path: str,
    preset: Annotated[
        str,
        typer.Option(
            '--preset',
            '-p',
            help='Which preset to use to create this package. Can be a named of an already installed preset, or an URI, in which case the preset will be downloaded.',
        ),
    ] = 'default',
    local: bool = typer.Option(
        False,
        '--local',
        '-l',
        help='Whether to inline the installed preset within the contest folder.',
    ),
):
    console.console.print(f'Creating new contest at [item]{path}[/item]...')

    fetch_info = get_preset_fetch_info(preset)
    if fetch_info is None:
        console.console.print(
            f'[error]Invalid preset name/URI [item]{preset}[/item].[/error]'
        )
        raise typer.Exit(1)

    if fetch_info.fetch_uri is not None:
        preset = presets.install_from_remote(fetch_info)

    preset_cfg = presets.get_installed_preset(preset)
    preset_path = (
        presets.get_preset_installation_path(preset)
        if preset_cfg.contest is not None
        else presets.get_preset_installation_path('default')
    )

    contest_path = (
        presets.get_preset_installation_path(preset) / preset_cfg.contest
        if preset_cfg.contest is not None
        else presets.get_preset_installation_path('default') / 'contest'
    )

    if not contest_path.is_dir():
        console.console.print(
            f'[error]Contest template [item]{contest_path}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)

    dest_path = pathlib.Path(path)

    if dest_path.exists():
        if not rich.prompt.Confirm.ask(
            f'Directory [item]{dest_path}[/item] already exists. Create contest in it? This might be destructive.',
            show_default=False,
            console=console.console,
        ):
            console.console.print(
                f'[error]Directory [item]{dest_path}[/item] already exists.[/error]'
            )
            raise typer.Exit(1)

    dest_path.mkdir(parents=True, exist_ok=True)
    shutil.copytree(str(contest_path), str(dest_path), dirs_exist_ok=True)
    shutil.rmtree(str(dest_path / 'build'), ignore_errors=True)
    shutil.rmtree(str(dest_path / '.box'), ignore_errors=True)
    shutil.rmtree(str(dest_path / '.local.rbx'), ignore_errors=True)
    # TODO: consider clearing build and .box recursively for nested problem directories
    for lock in dest_path.rglob('.preset-lock.yml'):
        lock.unlink(missing_ok=True)

    if local:
        shutil.copytree(str(preset_path), str(dest_path / '.local.rbx'))

    with utils.new_cd(dest_path):
        contest_utils.clear_all_caches()
        presets.generate_lock(preset if not local else presets.LOCAL)


@app.command('edit, e', help='Open contest.rbx.yml in your default editor.')
@within_contest
def edit():
    console.console.print('Opening contest definition in editor...')
    # Call this function just to raise exception in case we're no in
    # a problem package.
    find_contest()
    open_editor(find_contest_yaml() or pathlib.Path())


@app.command('add, a', help='Add new problem to contest.')
@within_contest
def add(path: str, short_name: str, preset: Optional[str] = None):
    name = pathlib.Path(path).stem
    utils.validate_field(ContestProblem, 'short_name', short_name)
    utils.validate_field(Package, 'name', name)

    if short_name in [p.short_name for p in find_contest_package_or_die().problems]:
        console.console.print(
            f'[error]Problem [item]{short_name}[/item] already exists in contest.[/error]',
        )
        raise typer.Exit(1)

    preset_lock = presets.get_preset_lock()
    if preset is None and preset_lock is not None:
        preset = preset_lock.preset_name
    creation.create(name, preset=preset, path=pathlib.Path(path))

    contest = find_contest_package_or_die()
    # Reassign mutable object before saving.
    contest.problems = sorted(
        [
            *contest.problems,
            ContestProblem(short_name=short_name, path=pathlib.Path(path)),
        ],
        key=lambda p: p.short_name,
    )

    save_contest(contest)
    console.console.print(
        f'Problem [item]{name} ({short_name})[/item] added to contest at [item]{path}[/item].'
    )


@app.command('remove, r', help='Remove problem from contest.')
@within_contest
def remove(path_or_short_name: str):
    contest = find_contest_package_or_die()

    kept_problems = []
    removed_problems = []
    for problem in contest.problems:
        if (
            problem.path == pathlib.Path(path_or_short_name)
            or problem.short_name == path_or_short_name
        ):
            removed_problems.append(problem)
        else:
            kept_problems.append(problem)

    contest.problems = kept_problems
    save_contest(contest)

    for problem in removed_problems:
        shutil.rmtree(str(problem.path), ignore_errors=True)
        console.console.print(
            f'Problem [item]{problem.short_name}[/item] removed from contest at [item]{problem.path}[/item].'
        )


@app.command(
    'each',
    help='Run a command for each problem in the contest.',
    context_settings={'allow_extra_args': True, 'ignore_unknown_options': True},
)
@within_contest
def each(ctx: typer.Context) -> None:
    command = ' '.join(['rbx'] + ctx.args)
    contest = find_contest_package_or_die()
    ok = True
    for problem in contest.problems:
        console.console.print(
            f'[status]Running [item]{command}[/item] for [item]{problem.short_name}[/item]...[/status]'
        )

        retcode = subprocess.call(
            command,
            cwd=problem.get_path(),
            shell=True,
        )
        ok = ok and retcode == 0
        console.console.print()

    if not ok:
        console.console.print(
            '[error]One of the commands above failed. Check the output![/error]'
        )
